'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

const NOP = () => {};

const forEach = (o, handler) => Object.keys(o).forEach(k => handler(o[k], k));

const __typePattern = /^\[object ([^\]]+)]$/;

const __getTypeOfObject = o => (__typePattern.exec(Object.prototype.toString.call(o)) || [undefined])[1];

const __serializeProp = (key, value, custom, serialized) => typeof value === 'string' ? `${key}="${value}"` : `${key}={${__serialize(value, custom, serialized)}}`;

const __serializeProps = (o, custom, serialized) => {
  const elemProps = o.props;
  const props = [];
  forEach(elemProps, (v, k) => {
    if (v !== undefined && k !== 'children') props.push(__serializeProp(k, v, custom, serialized));
  });
  if (o.key) props.push(__serializeProp('key', o.key, custom, serialized));
  if (o.ref) props.push(__serializeProp('ref', o.ref, custom, serialized));
  if (!props.length) return '';
  return ' ' + props.join(' ');
};

const __serializeChildren = (o, custom, serialized) => {
  const children = o.props.children;
  if (!children) return '';
  if (typeof children === 'string') return children;
  return Array.isArray(children) ? children.map(v => __serialize(v, custom, serialized)).join('') : __serialize(children, custom, serialized);
};

const __reactTypeNameReader = {
  String: type => type,
  Function: type => type.name,
  Symbol: type => Symbol.keyFor(type) === 'react.fragment' && 'Fragment' || undefined
};

const getTypeName = type => (__reactTypeNameReader[__getTypeOfObject(type)] || NOP)(type);

const __serializeReactElement = (o, custom, serialized) => {
  const type = getTypeName(o.type) || 'UNKNOWN';

  const children = __serializeChildren(o, custom, serialized);

  return `<${type}${__serializeProps(o, custom, serialized)}${children ? `>${children}</${type}>` : ' />'}`;
};

const __serializeReact = (o, custom, serialized) => {
  const symbolKey = Symbol.keyFor(o.$$typeof);

  switch (symbolKey) {
    case 'react.element':
      return __serializeReactElement(o, custom, serialized);

    default:
      return;
  }
};

const __serializeIfReact = (o, custom, serialized) => {
  if (Object.prototype.toString.call(o.$$typeof) === '[object Symbol]') {
    const key = Symbol.keyFor(o.$$typeof);
    if (key && key.indexOf('react.') === 0) return __serializeReact(o, custom, serialized);
  }
};

const __serializerByType = {
  BigInt: o => `${String(o)}n`,
  RegExp: o => `/${String(o)}/`,
  String: o => o.indexOf('"') === -1 && o.indexOf("'") !== -1 ? `"${o}"` : `'${o}'`,
  Function: o => o.name || 'Function',
  AsyncFunction: o => o.name || 'AsyncFunction',
  Date: o => `new Date(${Number(o)})`,
  Number: o => String(o),
  Boolean: o => String(o),
  Set: o => `new Set(${__serializeArray([...o.keys()], undefined, [])})`,
  Map: o => `new Map(${__serializeArray([...o.entries()], undefined, [])})`,
  Symbol: o => Symbol.keyFor(o) === undefined ? o.toString() : `Symbol.for('${Symbol.keyFor(o)}')`,
  Error: o => `new ${o.name}('${o.message}')`
};

const __serializeArray = (o, custom, serialized) => {
  const results = [];

  for (let i = 0; i < o.length; i++) {
    results.push(__serialize(o[i], custom, serialized));
  }

  return `[${results.join(', ')}]`;
};

const __serializeOptArray = (o, custom, serialized) => {
  if (Array.isArray(o)) return __serializeArray(o, custom, serialized);
};

const __serializeObject = (o, custom, serialized) => {
  const oKeys = Object.keys(o);
  oKeys.sort();
  const results = [];

  for (let i = 0; i < oKeys.length; i++) {
    const key = oKeys[i];
    results.push(`${key}: ${__serialize(o[key], custom, serialized)}`);
  }

  const objectType = o.constructor.name;
  const displayedType = objectType === 'Object' ? '' : objectType;
  return `${displayedType}{${results.join(', ')}}`;
};

const Serialize = {
  custom: (o, custom) => custom && custom(o),
  undefOrNull: o => o === undefined && 'undefined' || o === null && 'null' || undefined,
  flat: o => (__serializerByType[__getTypeOfObject(o)] || NOP)(o),
  cyclomatic: (o, _, serialized) => serialized.indexOf(o) !== -1 && '>CYCLOMATIC<' || undefined,
  react: __serializeIfReact,
  array: __serializeOptArray
};

const __serialize = (o, custom, serialized) => Serialize.custom(o, custom, serialized) || Serialize.undefOrNull(o, custom, serialized) || Serialize.flat(o, custom, serialized) || Serialize.cyclomatic(o, custom, serialized) || (nextSerialized => Serialize.react(o, custom, nextSerialized) || Serialize.array(o, custom, nextSerialized) || __serializeObject(o, custom, nextSerialized))([...serialized, o]);

const _serialize = (o, custom) => __serialize(o, custom, []);

const Serializer = {
  run: o => _serialize(o),
  create: custom => o => _serialize(o, custom)
};

const IGNORE = Symbol.for('__Spy_IGNORE__');
const serialize = Serializer.create(o => o === IGNORE && '>IGNORED<' || undefined);

const forEach$1 = (arrOrObj, handler) => {
  for (let key in arrOrObj) {
    if (arrOrObj.hasOwnProperty(key)) {
      handler(key, arrOrObj[key]);
    }
  }
};

const objectKeys = arrOrObj => {
  const keys = [];
  forEach$1(arrOrObj, key => keys.push(key));
  return keys;
};

const mergeArrays = (arr1, arr2) => {
  const result = [...arr1];
  forEach$1(arr2, (key, val) => {
    if (arr1.indexOf(val) === -1) {
      result.push(val);
    }
  });
  return result;
};

const SPY_COMPARE_FAILED = 'Spy.COMPARE failed';
const SPY_MAPPER_FAILED = 'Spy.MAPPER failed';

class SpyComparator {
  constructor(comparator) {
    _defineProperty(this, "_func", void 0);

    this._func = comparator;
  }

  compare(arg) {
    const result = this._func(arg);

    if (typeof result === 'string') return [`${SPY_MAPPER_FAILED} [${result}]`];
    if (result === false) return [SPY_COMPARE_FAILED];
  }

}

const COMPARE = comparator => new SpyComparator(comparator);

const MAPPER = (from, to) => new SpyComparator(mapper => {
  const result = mapper(...(Array.isArray(from) ? from : [from]));
  const diff = differenceOf(result, to);
  return diff ? `${serialize(result)} did not match ${serialize(to)}: ${diff}` : undefined;
});

const __different = type => ['different ' + type];

const __diff = (a, b, initial, useOwnEquals, alreadyComparedArray = []) => {
  if (a === IGNORE || b === IGNORE) return;
  if (a instanceof SpyComparator) return a.compare(b);
  if (b instanceof SpyComparator) return b.compare(a);
  if (a === b) return;
  if (a === undefined || b === undefined) return ['one was undefined'];
  if (a === null || b === null) return ['one was null'];
  const aClass = Object.prototype.toString.call(a);
  const bClass = Object.prototype.toString.call(b);
  if (aClass !== bClass) return __different(`object types: ${aClass} <-> ${bClass}`);

  switch (aClass) {
    case '[object RegExp]':
      if (String(a) === String(b)) {
        return;
      }

      return __different('regexp');

    case '[object String]':
      return __different('string');

    case '[object Function]':
      return __different('function');

    case '[object AsyncFunction]':
      return __different('async function');

    case '[object Number]':
      if (isNaN(a) && isNaN(b)) {
        return;
      }

      return __different('number');

    case '[object BigInt]':
      return __different('BigInt');

    case '[object Date]':
      if (Number(a) === Number(b)) {
        return;
      }

      return __different('date');

    case '[object Boolean]':
      return __different('bool');

    case '[object Symbol]':
      return __different('symbols');

    case '[object Error]':
      if (String(a) === String(b)) {
        return;
      }

      return __different('error');

    default:
      if (a.constructor !== b.constructor) {
        return __different('constructor');
      }

  }

  if (useOwnEquals && a.equals instanceof Function) {
    if (a.equals(b)) {
      return;
    }

    return ['own equals method failed <- ' + 'Maybe you want to disable the usage ' + 'of own equals implementation? ' + '[ Use: spy.configure({useOwnEquals: false}) ]'];
  }

  if (alreadyComparedArray.indexOf(a) !== -1) {
    return;
  }

  const compared = [...alreadyComparedArray, a];
  const keys = mergeArrays(objectKeys(a), objectKeys(b));

  for (let i = 0; i < keys.length; i++) {
    const key = keys[i];

    const diff = __diff(a[key], b[key], false, useOwnEquals, compared);

    if (diff !== undefined) {
      return [key, ...diff];
    }
  }
};

const __serializeDifferentProp = (obj, diff) => {
  let toSerialize = obj;
  diff.slice(0, -1).forEach(key => {
    toSerialize = toSerialize[key];
  });
  return serialize(toSerialize);
};

const __diffToStr = (diff, info = '') => {
  if (diff.length === 1) return diff[0];
  return `--> ${diff[0]} / ` + diff.slice(1).join(' / ') + info;
};

const differenceOf = (a, b, config = {
  useOwnEquals: true
}) => {
  const diff = __diff(a, b, true, config.useOwnEquals);

  if (!diff) return;

  const diffStr = __diffToStr(diff);

  if (diff.length < 2) return diffStr;

  const diffProp1 = __serializeDifferentProp(a, diff);

  const lastPart = diff[diff.length - 1];
  if (lastPart.indexOf(SPY_MAPPER_FAILED) === 0) return diffStr;
  const info = lastPart === SPY_COMPARE_FAILED ? `called with: ${diffProp1}` : `${diffProp1} != ${__serializeDifferentProp(b, diff)}`;
  return `${diffStr} [${info}]`;
};

const toError = (msgOrError, spyName) => msgOrError instanceof Error ? msgOrError : new Error(msgOrError || `${spyName} was requested to throw`);

const restoreAttributeForEntry = value => {
  const {
    obj,
    method,
    methodName
  } = value;

  if (obj) {
    obj[methodName] = method;
  }
};

function SpyRegistry() {
  if (!(this instanceof SpyRegistry)) {
    throw new Error('\n\nPlease make sure to use this constructor only with "new" keyword.\n\n');
  }

  this.register = {};
  this.persReg = {};
  this.registerCount = 0;
}

SpyRegistry.prototype.restoreAll = function () {
  forEach$1(this.register, (ignored, entry) => {
    restoreAttributeForEntry(entry);
  });
  this.register = {};
};

SpyRegistry.prototype.restore = function (index) {
  const entry = this.register[index];

  if (entry) {
    restoreAttributeForEntry(entry);
    delete this.register[index];
  }
};

SpyRegistry.prototype.push = function (obj, methodName) {
  this.registerCount += 1;
  this.register[this.registerCount] = {
    obj,
    method: obj[methodName],
    methodName
  };
  return this.registerCount;
};

SpyRegistry.prototype.getOriginalMethod = function (index) {
  const entry = this.register[index];

  if (entry) {
    return entry.method;
  }
};

SpyRegistry.prototype.persist = function (index, intoPersReg) {
  const fromReg = intoPersReg ? this.register : this.persReg;
  const toReg = intoPersReg ? this.persReg : this.register;
  const entry = fromReg[index];

  if (entry) {
    toReg[index] = entry;
    delete fromReg[index];
  }
};

const uninitialized = method => () => {
  throw new Error(`Method '${method}' was not initialized on Mock.`);
};

const defaultScope = Symbol('__Spy_global__');
const _mocks = {
  [defaultScope]: []
};
let scope = defaultScope;
const setScope = scoping => {
  if (scoping) {
    _mocks[scoping] = [];
    scope = scoping;
  } else scope = defaultScope;
};

const registerMock = (mocked, mock = {}) => {
  _mocks[scope].push({
    mocked,
    mock,
    scope
  });

  return mock;
};

const createMock = (obj, methods) => {
  const mock = registerMock(obj);
  forEach$1(methods, (_, method) => {
    mock[method] = uninitialized(method);
  });
  return mock;
};

const couldNotInitError = (scope, additional) => new Error(`Could not initialize mock for ${scope === defaultScope ? 'global scope' : `scope "${scope}"`}, because:\n${additional}`);

const initMock = ({
  mocked,
  mock,
  scope
}, spyOn) => {
  forEach$1(mock, method => {
    try {
      mock[method] = spyOn(mocked, method);
    } catch (e) {
      throw couldNotInitError(scope, e.message);
    }
  });
};

const initMockScope = (scoping, spyOn) => {
  forEach$1(_mocks[scoping], (_, mock) => initMock(mock, spyOn));
};

const initMocks = (spyOn, scoping) => {
  initMockScope(defaultScope, spyOn);
  scoping && initMockScope(scoping, spyOn);
};

const _testSuite = {
  isJest: !!jest,
  isCJS: !!require
};
const runner = {};
const oldDescribe = describe;

describe = (name, suite) => {
  oldDescribe(name, () => {
    setScope(name);
    beforeEach(() => {
      runner.beforeEach && runner.beforeEach(name);
    });
    afterEach(() => {
      runner.afterEach && runner.afterEach(name);
    });
    const rv = suite();
    setScope(undefined);
    return rv;
  });
};

const configure = other => {
  if (other.afterEach) runner.afterEach = other.afterEach;
  if (other.beforeEach) runner.beforeEach = other.beforeEach;
};

const addSnapshotSerializer = serializer => {
  _testSuite.isJest && expect && expect.addSnapshotSerializer && expect.addSnapshotSerializer(serializer);
};

const __caller = stackNum => {
  const traceFn = Error.prepareStackTrace;

  Error.prepareStackTrace = (err, stack) => stack;

  const stack = new Error().stack;
  Error.prepareStackTrace = traceFn;
  return stack[stackNum].getFileName();
};

const __callerBasedir = stackNum => require('path').dirname(__caller(stackNum));

const STACK_NUM_CREATE_MOCK = 4;

const createMock$1 = (SPY, moduleName, names) => {
  if (!_testSuite.isCJS) throw new Error('Spy.moduleMock works only if your test runner executes with CommonJS');
  return SPY.mock(require(require('path').join(__callerBasedir(STACK_NUM_CREATE_MOCK), moduleName)), ...names);
};

const TestSuite = {
  addSnapshotSerializer,
  createMock: createMock$1,
  configure
};

const registry = new SpyRegistry();
let __LOCK__ = true;
const Symbols = {
  name: Symbol('__Spy_name__'),
  snap: Symbol('__Spy_snap__'),
  isSpy: Symbol('__Spy_isSpy__'),
  func: Symbol('__Spy_func__'),
  calls: Symbol('__Spy_calls__'),
  config: Symbol('__Spy_config__'),
  index: Symbol('__Spy_index__')
};
TestSuite.addSnapshotSerializer({
  test: v => v && v[Symbols.isSpy],
  print: spy => spy[Symbols.snap]
});
const DefaultSettings = {
  useOwnEquals: true
};
const SpyFunctions = {
  configure(config) {
    if (config.useOwnEquals !== undefined) {
      this[Symbols.config].useOwnEquals = config.useOwnEquals;
    }

    if (config.persistent !== undefined) {
      if (!this[Symbols.index]) {
        throw new Error(`\n\n${this[Symbols.name]} can not` + ' be configured to be persistent!' + ' It does not mock any object.');
      }

      this[Symbols.config].persistent = config.persistent;
      registry.persist(this[Symbols.index], this[Symbols.config].persistent);
    }

    return this;
  },

  calls(...funcs) {
    if (funcs.length === 0) {
      this[Symbols.func] = () => {};

      return this;
    }

    const max = funcs.length - 1;
    let counter = -1;

    this[Symbols.func] = (...args) => {
      counter++;
      return funcs[max < counter ? max : counter](...args);
    };

    return this;
  },

  returns(...args) {
    return this.calls(...args.map(arg => () => arg));
  },

  resolves(...args) {
    return this.returns(...(args.length ? args : [undefined]).map(arg => Promise.resolve(arg)));
  },

  rejects(...msgOrErrors) {
    return this.calls(...(msgOrErrors.length ? msgOrErrors : [undefined]).map(msgOrError => () => Promise.reject(toError(msgOrError, this[Symbols.name]))));
  },

  throws(msgOrError) {
    this[Symbols.func] = () => {
      throw toError(msgOrError, this[Symbols.name]);
    };

    return this;
  },

  reset() {
    this[Symbols.calls] = [];
    return this;
  },

  restore() {
    if (this[Symbols.config].persistent) {
      throw new Error(`\n\n${this[Symbols.name]} can not be restored!` + ' It was configured to be persistent.');
    }

    registry.restore(this[Symbols.index]);
    return this;
  },

  transparent() {
    return this.transparentAfter(0);
  },

  transparentAfter(callCount) {
    const oldFunc = this[Symbols.func];

    this[Symbols.func] = (...args) => {
      if (this[Symbols.calls].length > callCount) {
        const originalMethod = registry.getOriginalMethod(this[Symbols.index]);

        if (originalMethod) {
          return originalMethod(...args);
        }

        return;
      }

      return oldFunc(...args);
    };

    return this;
  },

  wasCalled(callCount) {
    const madeCalls = this[Symbols.calls].length;

    if (typeof callCount === 'number') {
      if (madeCalls !== callCount) {
        throw new Error(`\n\n${this[Symbols.name]} was called ${madeCalls} times,` + ` but there were expected ${callCount} calls.\n\n` + 'Actually there were:\n\n' + this.showCallArguments());
      }
    } else if (madeCalls === 0) {
      throw new Error(`\n\n${this[Symbols.name]} was never called!\n\n`);
    }
  },

  hasCallHistory(...callHistory) {
    const madeCalls = this[Symbols.calls];
    const callCount = callHistory.length;

    if (madeCalls.length !== callCount) {
      throw new Error(`\n\n${this[Symbols.name]} was called ${madeCalls.length} times,` + ` but the expected call history includes exactly ${callHistory.length} calls.\n\n` + 'Actually there were:\n\n' + this.showCallArguments());
    }

    const modifiedCallHistory = callHistory.map(arg => Array.isArray(arg) ? arg : [arg]);
    let hasErrors = false;
    const diffInfo = madeCalls.map((call, index) => {
      const diff = differenceOf(call.args, modifiedCallHistory[index], this[Symbols.config]);
      if (diff) hasErrors = true;
      return diff;
    });
    if (hasErrors) throw new Error(`\n\n${this[Symbols.name]} was considered` + ' to be called with the following arguments in the given order:\n\n' + `${modifiedCallHistory.map((entry, index) => `call ${index}: ${serialize(entry)}`).join('\n')}\n\n` + 'Actually there were:\n\n' + this.showCallArguments(diffInfo));
  },

  wasNotCalled() {
    const madeCalls = this[Symbols.calls];

    if (madeCalls.length !== 0) {
      throw new Error(`\n\n${this[Symbols.name]} was not` + ' considered to be called.\n\n' + 'Actually there were:\n\n' + this.showCallArguments());
    }
  },

  wasCalledWith(...args) {
    const madeCalls = this[Symbols.calls];

    if (madeCalls.length === 0) {
      throw new Error(`\n\n${this[Symbols.name]} was never called!\n\n`);
    }

    const diffInfo = [];

    for (let i = 0; i < madeCalls.length; i++) {
      const diff = differenceOf(madeCalls[i].args, args, this[Symbols.config]);

      if (!diff) {
        return;
      }

      diffInfo.push(diff);
    }

    throw new Error(`\n\n${this[Symbols.name]} was considered` + ' to be called with the following arguments:\n\n' + `    --> ${serialize(args)}\n\n` + 'Actually there were:\n\n' + this.showCallArguments(diffInfo));
  },

  wasNotCalledWith(...args) {
    let errorOccurred = false;

    try {
      this.wasCalledWith(...args);
    } catch (e) {
      errorOccurred = true;
    }

    if (!errorOccurred) {
      throw new Error(`\n\n${this[Symbols.name]} was called` + ' unexpectedly with the following arguments:\n\n' + `    --> ${serialize(args)}\n\n`);
    }
  },

  getCallArguments(callNr = 0) {
    const madeCalls = this[Symbols.calls];

    if (callNr % 1 !== 0 || callNr >= madeCalls.length) {
      throw new Error(`\n\nThe provided callNr "${callNr}" was not valid.\n\n` + `Made calls for ${this[Symbols.name]}:\n\n` + this.showCallArguments());
    }

    return madeCalls[callNr].args;
  },

  getCallArgument(callNr = 0, argNr = 0) {
    return this.getCallArguments(callNr)[argNr];
  },

  getCallCount() {
    return this[Symbols.calls].length;
  },

  showCallArguments(additionalInformation = []) {
    const madeCalls = this[Symbols.calls];

    if (madeCalls.length === 0) {
      return `${this[Symbols.name]} was never called!\n`;
    }

    let response = '';

    for (let i = 0; i < madeCalls.length; i++) {
      const args = serialize(madeCalls[i].args);
      response += `call ${i}: ${args}\n`;

      if (additionalInformation[i]) {
        response += `        ${additionalInformation[i]}\n`;
      }
    }

    return response;
  }

};
const AllCreatedSpies = [];

class Spy {
  constructor(name = '', __mock) {
    const spy = function (...args) {
      spy[Symbols.calls].push({
        args
      });
      return spy[Symbols.func](...args);
    };

    if (__mock && !__LOCK__) {
      spy[Symbols.index] = registry.push(__mock.obj, __mock.methodName);
      spy[Symbols.name] = `the spy on '${name}'`;
      spy[Symbols.snap] = `Spy.on(${name})`;
    } else {
      spy[Symbols.index] = null;
      spy[Symbols.name] = name || 'the spy';
      spy[Symbols.snap] = `Spy(${name})`;
    }

    spy[Symbols.isSpy] = true;

    spy[Symbols.func] = () => {};

    spy[Symbols.calls] = [];
    spy[Symbols.config] = {
      useOwnEquals: DefaultSettings.useOwnEquals
    };
    forEach$1(SpyFunctions, (key, value) => {
      spy[key] = value;
    });
    AllCreatedSpies.push(spy);
    return spy;
  }

  static configure(config) {
    if (config.useOwnEquals !== undefined) {
      DefaultSettings.useOwnEquals = config.useOwnEquals;
    }

    TestSuite.configure({
      afterEach: config.afterEach,
      beforeEach: config.beforeEach
    });
  }

  static on(obj, methodName) {
    const method = obj[methodName];

    if (!(typeof method === 'function')) {
      throw new Error(`The object attribute '${methodName}' ` + `was: ${serialize(method)}\n\n` + 'You should only spy on functions!');
    }

    if (method[Symbols.isSpy]) {
      throw new Error(`The objects attribute '${methodName}'` + ' was already spied. Please make sure to spy' + ' only once at a time at any attribute.');
    }

    __LOCK__ = false;
    const spy = new Spy(methodName, {
      obj,
      methodName
    });
    __LOCK__ = true;
    obj[methodName] = spy;
    return spy;
  }

  static mock(obj, ...methodNames) {
    return createMock(obj, methodNames);
  }

  static mockModule(moduleName, ...methodNames) {
    return TestSuite.createMock(Spy, moduleName, methodNames);
  }

  static initMocks(scope) {
    initMocks(Spy.on, scope);
  }

  static restoreAll() {
    registry.restoreAll();
  }

  static resetAll() {
    AllCreatedSpies.forEach(spy => spy.reset());
  }

}

_defineProperty(Spy, "IGNORE", IGNORE);

_defineProperty(Spy, "COMPARE", COMPARE);

_defineProperty(Spy, "MAPPER", MAPPER);

const defaultHooks = {
  beforeEach: Spy.initMocks,
  afterEach: () => {
    Spy.restoreAll();
    Spy.resetAll();
  }
};
TestSuite.configure(defaultHooks);

exports.Spy = Spy;
